/*
 * lcd_i2c.c
 *
 *  Created on: 23.03.2019
 *      Author: harry
 */

#include "i2c.h"

#include "lcd_config.h"
#include "lcd_map.inc"
#include "lcd_i2c.h"

/**********************************************************************************
 * there is usually no need to change anything below this line
 *
 * Constants and usefull-macros
 **********************************************************************************/

#define RS_Pin	0x0
#define RW_Pin	0x01
#define	E_PIN	0x02
#define BL_PIN	0x03

#if PCF8574_TYPE == PCF8574
#define PCF8574_ADR(x)	((0x20+x)<<1)
#else
#if PCF8574_TYPE == PCF8574A
#define PCF8574_ADR(x)	((0x38+x)<<1)
#endif
#endif

#define LCD_CLEARDISPLAY 	0x01
#define LCD_RETURNHOME 		0x02
#define LCD_DISPLAYON       0x0C
#define LCD_SETCGRAMADDR	0x40
#define LCD_SETDDRAMADDR    0x80

#define HIGH_NIBBLE(x) ((uint8_t)((x) & 0xf0))
#define LOW_NIBBLE(x)  ((uint8_t)((x) << 4))

/**********************************************************************************
* local variables
*
**********************************************************************************/

static uint8_t port_state = 0;
static uint8_t lx, ly, lp;
static I2C_HandleTypeDef *hlcdp;

/**********************************************************************************
 * private functions
 *
 **********************************************************************************/

static void send_nibble(uint8_t nibble)
{
	uint8_t cmd[2];
	port_state = (port_state & 0x0f) | nibble;
	cmd[0] = port_state | (1 << E_PIN);
	cmd[1] = port_state;
	HAL_I2C_Master_Transmit(hlcdp, PCF8574_ADR(LCD_ADR), cmd, 2, 2);
}

static void send_byte(uint8_t byte)
{
	uint8_t cmd[4];

	port_state = ((port_state & 0x0f) | HIGH_NIBBLE(byte));
	cmd[0] = port_state | (1 << E_PIN);
	cmd[1] = port_state;
	port_state = ((port_state & 0x0f) | LOW_NIBBLE(byte));
	cmd[2] = port_state | (1 << E_PIN);
	cmd[3] = port_state;
	HAL_I2C_Master_Transmit(hlcdp, PCF8574_ADR(LCD_ADR), cmd, 4, 4);
}

static void send_cmd(uint8_t cmd)
{
	port_state &= ~(1 << RS_Pin);	// Clear RS
	send_byte(cmd);
}

static void send_dta(uint8_t cmd)
{
	port_state |= (1 << RS_Pin);	// set RS
	send_byte(cmd);
}

/**********************************************************************************
 * public functions
 *
 **********************************************************************************/

/**********************************************************************************
 * Init
 * hlcd is a pointer to the HAL-i²c-handle in use
 * example:
 * lcd_Init(&hi2c1);
 *
 **********************************************************************************/

void lcd_Init(I2C_HandleTypeDef *hlcd)
{
	hlcdp = hlcd;
	HAL_Delay(350);
	send_nibble(0x30);
	HAL_Delay(40);
	send_nibble(0x30);
	HAL_Delay(2);
	send_nibble(0x30);
	HAL_Delay(2);
	send_nibble(0x20);
	send_cmd(LCD_FUNCTIONSET);
	send_cmd(LCD_ENTRYMODESET);
	send_cmd(LCD_DISPLAYON);

	port_state |= (1 << BL_PIN);		// enable Backlight by default
	lcd_clear();
}

/**********************************************************************************
 * clear whole lcd
 *
 **********************************************************************************/

void lcd_clear(void)
{
	send_cmd(LCD_CLEARDISPLAY);
	lx = ly = 0;
	lp = lcd_map[ly][lx];
	HAL_Delay(2);		// required! - see datasheet page 24
}

/**********************************************************************************
 * set cursor to {0|0}
 *
 **********************************************************************************/

void lcd_home(void)
{
	lx = ly = 0;
	lp = lcd_map[ly][lx];
	send_cmd(lp | LCD_SETDDRAMADDR);
}

/**********************************************************************************
 * set cursor to {x|y}
 *
 **********************************************************************************/

void lcd_gotoXY(uint8_t x, uint8_t y)
{
	if ((y < LCD_LINES) && (x << LCD_COLUMNS))
	{
		lx = x;
		ly = y;
		lp = lcd_map[y][x];
		send_cmd(lp | LCD_SETDDRAMADDR);
	}
}

/**********************************************************************************
 * switch backlight on or off
 * example:
 * lcd_backlight(TRUE);
 *
 **********************************************************************************/

void lcd_backlight(uint8_t Light)
{
	port_state = (port_state & ~(1 << BL_PIN)) | ((Light & 1) << BL_PIN);
	HAL_I2C_Master_Transmit(hlcdp, PCF8574_ADR(LCD_ADR), &port_state, 1, 1);
}


/**********************************************************************************
 * Todo: define custom chars
 *
 **********************************************************************************/

void lcd_define_char(uint8_t n_char, uint8_t *data)
{
	if (LCD_FUNCTIONSET & 0x04)
	{

	}
	else
	{

	}
}

/**********************************************************************************
 * write a single char at cursor-position
 * at EOL wrap to next line
 * at EOS wrap to home
 *
 **********************************************************************************/
void lcd_putc(char c)
{
	if (lcd_map[ly][lx] != ++lp)
	{
		lp = lcd_map[ly][lx];
		send_cmd(lp | LCD_SETDDRAMADDR);
	}
	send_dta(c);
	if (++lx == LCD_COLUMNS)
	{
		lx = 0;
		if (++ly == LCD_LINES)
			ly = 0;
	}
}

/**********************************************************************************
 * write a whole string at cursor-position
 *
 **********************************************************************************/

void lcd_puts(char *s)
{
	for (uint8_t i = 0; s[i] != '\0'; i++)
		lcd_putc(s[i]);
}
